/*-
 * This file contain code from FreeBSD's
 * /usr/src/usr.sbin/acpi/acpiconf/acpiconf.c.
 *
 * Copyright (c) 1999 Mitsuru IWASAKI <iwasaki@FreeBSD.org>
 * Copyright (c) 2009 Marc Lagrange <rhaamo@gruik.at>
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: acpiconf.c,v 1.5 2000/08/08 14:12:19 iwasaki Exp $
 *	$FreeBSD: releng/8.0/usr.sbin/acpi/acpiconf/acpiconf.c 193531 2009-06-05 18:50:45Z jkim $
 */

#include "hwstat.h"

#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/sysctl.h>

#include <dev/acpica/acpiio.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <sysexits.h>	/* used for EX_OSFILE & cie */
#include <unistd.h>


extern unsigned int flags;

static int	acpifd = -1;

/** ACPI Initialisation
 */
static void
acpi_init(void)
{
	acpifd = open(ACPIDEV, O_RDWR);
	if (acpifd == -1)
		acpifd = open(ACPIDEV, O_RDONLY);
	if (acpifd == -1)
		err(EX_OSFILE, ACPIDEV);
}


enum acpi_batterr {
	ACPI_BATTERR_OK, /* = 0 */
	ACPI_BATTERR_INVALID_ID,
	ACPI_BATTERR_GET_BATTINFO,
	ACPI_BATTERR_NOT_PRESENT,
	ACPI_BATTERR_GET_BIF,
	ACPI_BATTERR_GET_BST,
};
/** Get ACPI Battery informations
 * \param int num
 * \return enum acpi_batterr
 */
static enum acpi_batterr
acpi_battinfo(int num)
{
	printf("\n[battery%i]\n", num);

	union acpi_battery_ioctl_arg battio;
	const char *pwr_units;
	int hours, min;

	if (num < 0 || num > 64)
		return (ACPI_BATTERR_INVALID_ID);

	/* Test if battery is present. */
	battio.unit = num;
	if (ioctl(acpifd, ACPIIO_BATT_GET_BATTINFO, &battio) == -1)
		return (ACPI_BATTERR_GET_BATTINFO);
	if (battio.battinfo.state == ACPI_BATT_STAT_NOT_PRESENT)
		return (ACPI_BATTERR_NOT_PRESENT);
	else
		printf("\tPresent:\t\t\tyes\n");

	/* Print battery design information. */
	battio.unit = num;
	if (ioctl(acpifd, ACPIIO_BATT_GET_BIF, &battio) == -1)
		return (ACPI_BATTERR_GET_BIF);

	if (battio.bif.units == 0)
		pwr_units = "mW";
	else
		pwr_units = "mA";

	if (battio.bif.dcap == UNKNOWN_CAP)
		printf("\tDesign capacity:\t\tunknown\n");
	else {
		printf("\tDesign capacity:\t\t%d\t\t\t\t%sh\n", battio.bif.dcap,
		    pwr_units);
	}
	if (battio.bif.lfcap == UNKNOWN_CAP)
		printf("\tLast full capacity:\t\tunknown\n");
	else {
		printf("\tLast full capacity:\t\t%d\t\t\t\t%sh\n", battio.bif.lfcap,
		    pwr_units);
	}
	printf("\tTechnology:\t\t\t%s\n", battio.bif.btech == 0 ?
	    "primary (non-rechargeable)" : "secondary (rechargeable)");
	if (battio.bif.dvol == UNKNOWN_CAP)
		printf("\tDesign voltage\t\tunknown\n");
	else
		printf("\tDesign voltage\t\t\t%d\t\t\t\tmV\n", battio.bif.dvol);
	printf("\tCapacity (warn):\t\t%d\t\t\t\t%sh\n", battio.bif.wcap, pwr_units);
	printf("\tCapacity (low):\t\t\t%d\t\t\t\t%sh\n", battio.bif.lcap, pwr_units);
	printf("\tLow/warn granularity:\t\t%d\t\t\t\t%sh\n", battio.bif.gra1, pwr_units);
	printf("\tWarn/full granularity:\t\t%d\t\t\t\t%sh\n", battio.bif.gra2, pwr_units);
	printf("\tModel number:\t\t\t%s\n", battio.bif.model);
	printf("\tSerial number:\t\t\t%s\n", battio.bif.serial);
	printf("\tType:\t\t\t\t%s\n", battio.bif.type);
	printf("\tOEM info:\t\t\t%s\n", battio.bif.oeminfo);

	/* Print current battery state information. */
	battio.unit = num;
	if (ioctl(acpifd, ACPIIO_BATT_GET_BATTINFO, &battio) == -1)
		return (ACPI_BATTERR_GET_BATTINFO);
	printf("\tState:\t\t\t\t");
	if (battio.battinfo.state == 0)
		printf("high");
	if (battio.battinfo.state & ACPI_BATT_STAT_CRITICAL)
		printf("critical");
	if (battio.battinfo.state & ACPI_BATT_STAT_DISCHARG)
		printf("discharging");
	if (battio.battinfo.state & ACPI_BATT_STAT_CHARGING)
		printf("charging");
	printf("\n");
	if (battio.battinfo.cap == -1)
		printf("\tRemaining capacity:\t\tunknown\n");
	else
		printf("\tRemaining capacity:\t\t%d\t\t\t\tpercent\n",
				battio.battinfo.cap);
	if (battio.battinfo.min == -1)
		printf("\tRemaining time:\t\t\tunknown\n");
	else {
		hours = battio.battinfo.min / 60;
		min = battio.battinfo.min % 60;
		printf("\tRemaining time:\t\t\t%d:%02d\n", hours, min);
	}
	if (battio.battinfo.rate == -1)
		printf("\tPresent rate:\t\t\tunknown\n");
	else {
		printf("\tPresent rate:\t\t\t%d\t\t\t\t%s\n",
		    battio.battinfo.rate, pwr_units);
	}

	/* Print battery voltage information. */
	battio.unit = num;
	if (ioctl(acpifd, ACPIIO_BATT_GET_BST, &battio) == -1)
		return (ACPI_BATTERR_GET_BST);
	if (battio.bst.state != ACPI_BATT_STAT_NOT_PRESENT) {
		if (battio.bst.volt == UNKNOWN_VOLTAGE)
			printf("\tVoltage:\t\t\tunknown\n");
		else
			printf("\tVoltage:\t\t\t%d\t\t\t\tmV\n", battio.bst.volt);
	}

	return (ACPI_BATTERR_OK);

}

/* Main ACPI Batter function
 */
void
extended_acpi_battery(void)
{
	int battery_units, i, tmpi;
	char *name;

	acpi_init();
	/* C0DE */

	/* Verify if we have a ``hw.acpi.battery'' node */
	sysctlbyname("hw.acpi.battery", NULL, NULL, NULL, 0);
	if (errno != EISDIR) {
		tmpi = asprintf(&name,
		    "\n[ACPI Battery (%s)]\n\tNo ACPI Battery support.\n", ACPIDEV);
		if (tmpi < 0)
			err(ENOMEM, "asprintf");
		debug_print(name, flags);
	} else {
		battery_units = sysctl_get_int("hw.acpi.battery.units");
		if (battery_units < 0)
			printf("'hw.acpi.battery.units' returned -1, weird..\n");

		for (i = 0; i < battery_units; ++i) {
			switch (acpi_battinfo(i)) {
			case ACPI_BATTERR_OK:
				/* ok, do nada */
				break;
			case ACPI_BATTERR_INVALID_ID:
				printf("\tinvalid battery %i\n", i);
				break;
			case ACPI_BATTERR_GET_BATTINFO:
				printf("\tget battery user info (%i) failed\n", i);
				break;
			case ACPI_BATTERR_NOT_PRESENT:
				printf("\tBattery not present\n");
				break;
			case ACPI_BATTERR_GET_BIF:
				printf("\tget battery info (%i) failed\n", i);
				break;
			case ACPI_BATTERR_GET_BST:
				printf("\tget battery status (%i) failed\n", i);
				break;
			}
		}
	}

	/* E_C0DE */
	close(acpifd);
}
