/*
 * Copyright (c) 2009 Marc Lagrange <rhaamo@gruik.at>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "hwstat.h"

#include <sys/types.h>
#include <sys/sysctl.h>

#include <assert.h>
#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>


unsigned int flags;

/** Show the command usage
*/
void
usage(void)
{

	(void)fprintf(stderr, "usage: %s [-hvdcf]\n", getprogname());
	(void)fprintf(stderr, "\t-h\t\tShow this page\n");
	(void)fprintf(stderr, "\t-v\t\tShow hwstat version\n");
	(void)fprintf(stderr, "\t-d\t\tDisplay all info and non-working devices\n");
	(void)fprintf(stderr, "\t-f\t\tDisplay temperature in Fahrenheit\n");
	(void)fprintf(stderr, "\t-k\t\tDisplay the temperature in Kelvins\n");
	exit(EXIT_FAILURE);
}


/** Return the CPU Temperature for a specific CPU
 * The return value is in Kelvin
 * \param int cpu_number
 * \return int
 */
int
get_cpu_temp(int cpu_number)
{
	int	i, temp;
	size_t	len = sizeof(temp);
	char	*mibname;

	/* Get the CPU Temperature via dev.cpu.<number>.temperature */
	i = asprintf(&mibname, "dev.cpu.%i.temperature", cpu_number);
	if (i < 0)
		err(ENOMEM, "asprintf");
	sysctlbyname(mibname, &temp, &len, NULL, 0);
	free(mibname);

	return (temp);
}


/** Return the converted Temperature.
 * The return value is in Fahrenheit, Celsius or Kelvin
 * \param int temp
 * \return double
 */
double
convert_temp(int temp)
{
	double t;

	/* We convert temperature in double
	 * At the origin the value is like
	 * '1234' and we need '123.4' that is
	 * the real kelvin value.
	 */
	double kelvin = temp / 10.0;

	if (flags & HWSTAT_FFLAG) /* Show in Fahrenheit */
		t = KELV_TO_FAR(kelvin);
	else if (flags & HWSTAT_KFLAG) /* Show in Kelvin */
		t = kelvin;
	else /* Default, show in Celcius */
		t = KELV_TO_CEL(kelvin);
	return (t);
}


/** Return the Temperature extension.
 * \return char
 */
char
get_temp_ext(void)
{
	char c;

	if (flags & HWSTAT_FFLAG)
		c = 'F';
	else if (flags & HWSTAT_KFLAG)
		c = 'K';
	else
		c = 'C';
	return (c);
}


/** Print CPU temp line
 * \param int cpu_number
 * \return void
 */
void
print_cpu_temp(int cpu_number)
{
	/* Test if the Sysctl OID realy exist */
	int	i;
	char	*mibname;

	i = asprintf(&mibname, "dev.cpu.%i.temperature", cpu_number);
	if (i < 0)
		err(ENOMEM, "asprintf");
	if (sysctlbyname(mibname, NULL, NULL, NULL, 0) < 0) {
		/* Errr, something wrong... */
		(void)printf("\tCPU%i:\t\t\t\tCannot get temperature\n",
		    cpu_number);
	} else {
		int temp = get_cpu_temp(cpu_number);
#if defined(__FreeBSD__) && (__FreeBSD__ < 8)
		/* We need normalizing datas.
	 	 * FreeBSD < 8 have his dev.cpu.N.temperature
	 	 * in Celsius instead of Kelvin * 10.
	 	 */
		temp = CEL_TO_KEL(temp) * 10;
#endif
		(void)printf("\tCPU%i:\t\t\t\t%.1lf\t\t\t\t%c\n",
		    cpu_number,
		    convert_temp(temp),
		    get_temp_ext());
	}

	free(mibname);
}


/** Display CPUs temperature
 */
void
cpu_temp(void)
{
	/* retrieve the number of CPUs via hw.ncpu (HW_NCPU)
	 * via CTL_HW level
	 */
	int	mib[2], i, ncpu, error;
	size_t	len = sizeof(ncpu);

	mib[0] = CTL_HW;
	mib[1] = HW_NCPU;
	error = sysctl(mib, 2, &ncpu, &len, NULL, 0);
	assert(len = sizeof(ncpu));
	if (error < 0)
		(void)debug_print("\n[Coretemp]\n\tCan't get number of cpus.", flags);
	else {
		/* we have the number of CPUs in `ncpu' */

		(void)printf("[Coretemp]\n");

		for(i = 0; i < ncpu; ++i)
			print_cpu_temp(i);
	}
}


/** Function that get ACPI Battery infos
 */
void
acpi_battery(void)
{
	int battery_life, battery_time, battery_state, battery_units;

	/* Verify if we have a `hw.acpi.battery' node */
	sysctlbyname("hw.acpi.battery", NULL, NULL, NULL, 0);
	if (errno != EISDIR)
		debug_print("\n[ACPI Battery (sysctl)]\n\tNo ACPI Battery support.\n", flags);
	else {
		/* We have the node */
		(void)printf("\n[ACPI Battery (sysctl)]\n");
		/* hw.acpi.battery.life: percent capacity remaining */
		battery_life = sysctl_get_int("hw.acpi.battery.life");
		if (battery_life < 0)
			printf("\tCap remain:\t\t\t0\t\t\t\tpercents\n");
		else {
			printf("\tCap remain:\t\t\t%i\t\t\t\tpercents\n", battery_life);
		}
		/* hw.acpi.battery.time: remaining time in minutes */
		battery_time = sysctl_get_int("hw.acpi.battery.time");
		if (battery_time < 0) {
			if (sysctl_get_int("hw.acpi.acline"))
				(void)printf("\tTime remain:\t\t\tOn AC\n");
			else
				(void)printf("\tTime remain:\t\t\tN/A\n");
		} else {
			printf("\tLife time:\t\t\t%i\t\t\t\tMin(s)\n", battery_time);
		}
		/* hw.acpi.battery.state: current status flags */
		battery_state = sysctl_get_int("hw.acpi.battery.state");
		printf("\tState:\t\t\t\t%i\n", battery_state);
		/* hw.acpi.battery.units: number of batteries */
		battery_units = sysctl_get_int("hw.acpi.battery.units");
		if (battery_units < 0)
			printf("\tUnits:\t\t\t\t-1 returned, somwthing weird.\n");
		else {
			printf("\tUnits:\t\t\t\t%i\n", battery_units);
		}
	}

}


/** Function that get ACPI Temperature
 */
void
acpi_temp(void)
{
	/* get all hw.acpi.thermal.tzN.temperature and
	 * print temp for each
	 */
	int	i;
	int	loop = 1;
	int	acpi_number = 0;
	char	*mibname;

	do {
		int	temp, error, sensor_found = 0;
		size_t	len = sizeof(temp);

		i = asprintf(&mibname, "hw.acpi.thermal.tz%i.temperature", (int)acpi_number);
		if (i < 0)
			err(ENOMEM, "asprintf");
		error = sysctlbyname(mibname, &temp, &len, NULL, 0);
		free(mibname);
		if (error < 0) {
			/* Error, possibility of non-existant node, set acpi_number to NULL */
			if (acpi_number == 0)
				debug_print("\n[ACPI Thermal]\n\tNo ACPI Thermal Zone Support.\n", flags);
			loop = 0;
		} else {
			if (sensor_found == 0) {
				(void)printf("\n[ACPI Thermal]\n");
				sensor_found = 1;
			}
			printf("\ttz%i:\t\t\t\t%.1lf\t\t\t\t%c\n",
			    (int)acpi_number,
			    convert_temp(temp),
			    get_temp_ext());
		}
		/* Increment to try the next Tz zome */
		++acpi_number;
	} while (loop);
}


/** Function for testing flags integrity
 */
void
test_flags_integrity(void)
{

	/* Constraint violation :
	 *  fahrenheit && celsius / HWSTAT_FFLAG && HWSTAT_KFLAG
	 */
	if ((flags & HWSTAT_FFLAG) && (flags & HWSTAT_KFLAG)) {
		(void)printf("You can't choice to show temperature un Fahrenheit and Kelvin\n");
		exit(EXIT_FAILURE);
	}
}


/** Function for getting *all* infos
 */
void
get_all_infos(void)
{

	printf("\t\t\t\t\tCurrent\t\t\t\tUnit\n");

	/* CPU Temperature via coretemp(4) */
	cpu_temp();

	/* ACPI Temperature via acpi_thermal(4) */
	acpi_temp();

	/* ACPI Battery via sysctl */
	acpi_battery();

	/* Extended ACPI Battery infos using acpi(4) */
	extended_acpi_battery();

}


/** Main function
 * \param argc ?
 * \param argv ?
 * \return 0
 */
int
main(int argc, char **argv)
{
	int 		i;
	unsigned int 	interval = 0;
	char		*endptr;

	if (argc < 2)
		flags = HWSTAT_CFLAG;

	setprogname(argv[0]);

	while ((i = getopt(argc, argv, "hvdfki:")) != -1) {
		switch (i) {
		case 'v':           /* print version */
			printf("hwstat "HWSTAT_VERSION"\n");
			exit(EXIT_SUCCESS);
			break;

		case 'd':           /* Show all 'debug' infos */
			flags |= HWSTAT_DFLAG;
			break;

		case 'f':           /* display temperature in Farenheit */
			flags |= HWSTAT_FFLAG;
			break;

		case 'k':           /* display temperature in Kelvin */
			flags |= HWSTAT_KFLAG;
			break;
		case 'i':
			flags |= HWSTAT_IFLAG;
			interval = (unsigned int)strtoul(optarg, &endptr, 10);
			if (endptr == optarg || *endptr != '\0')
				err(EINVAL, "bad interval value: %s", optarg);
			break;
		case 'h': /* FALLTHROUGH */
		default:  /* show help */
			usage();
			/* NOTREACHED */
		}
	}

	/*
	argc -= optind;
	argv += optind;
	*/


	test_flags_integrity();

	if (flags & HWSTAT_IFLAG) {
		for (;;) {
			get_all_infos();
			(void)fflush(stdout);
			(void)sleep(interval);
		}
	} else
		get_all_infos();

	return (EXIT_SUCCESS);
}
